package geniusweb.domaineditor.model;

import java.math.BigDecimal;
import java.util.HashMap;
import java.util.Map;

import geniusweb.issuevalue.NumberValueSet;
import geniusweb.issuevalue.ValueSet;
import tudelft.utilities.immutablelist.Range;
import tudelft.utilities.listener.DefaultListenable;
import tudelft.utilities.logging.Reporter;
import tudelft.utilities.mvc.model.BasicModel;
import tudelft.utilities.mvc.model.ListModel;
import tudelft.utilities.mvc.model.NumberModel;
import tudelft.utilities.mvc.model.events.Changed;
import tudelft.utilities.mvc.model.events.Event;

/**
 * Model for a {@link NumberValueSet}
 *
 */
public class NumberValueSetModel extends DefaultListenable<Event>
		implements ValueSetModel {

	private final NumberModel low, high, step;
	protected final Reporter log;

	public NumberValueSetModel(NumberModel lowv, NumberModel highv,
			NumberModel stepv, Reporter log) {
		this.low = lowv;
		this.high = highv;
		this.step = stepv;
		this.log = log;

		check();

		// 'this' refers to surrounding class inside lambda.
		low.addListener((e) -> check());
		high.addListener((e) -> check());
		step.addListener((e) -> check());
	}

	private void check() throws IllegalStateException {
		if (low.getValue().compareTo(high.getValue()) > 0)
			throw new IllegalStateException("low must be <= high");
		if (step.getValue().compareTo(BigDecimal.ZERO) <= 0)
			throw new IllegalStateException("step must be >0");
		notifyListeners(new Changed(this, null));
	}

	/*
	 * Default, bit arbitrary values.
	 */
	public NumberValueSetModel(Reporter log) {
		this(new NumberModel(BigDecimal.ZERO, log),
				new NumberModel(BigDecimal.TEN, log),
				new NumberModel(BigDecimal.ONE, log), log);
	}

	@Override
	public String toString() {
		return "(" + low + high + step + ")";
	}

	public NumberModel getLow() {
		return low;
	}

	public NumberModel getHigh() {
		return high;

	}

	public NumberModel getStep() {
		return step;
	}

	@Override
	public BasicModel first() {
		return low;
	}

	@Override
	public ListModel getList() {
		/**
		 * Return a ListModel backed up by the original set. It contains the
		 * possible values in the range.
		 */
		return new MyListModel(this);
	}

	@Override
	public NumberValueSet getCurrentValue() throws IllegalStateException {
		return new NumberValueSet(
				new Range(low.getValue(), high.getValue(), step.getValue()));
	}

	@Override
	public void setCurrentValue(ValueSet obj) throws IllegalArgumentException {
		if (!(obj instanceof NumberValueSet))
			throw new IllegalArgumentException(
					"Argument must be a NumberValueSet but got " + obj);
		NumberValueSet nv = (NumberValueSet) obj;
		// order of change is important, avoiding illegal ranges:
		// if low must be increased, change high first.
		// if high must be decreased, change low first.
		// otherwise the change order is irrelevant.
		if (nv.getRange().getLow().compareTo(low.getValue()) > 0) {
			high.setValue(nv.getRange().getHigh());
			low.setValue(nv.getRange().getLow());
		} else {
			low.setValue(nv.getRange().getLow());
			high.setValue(nv.getRange().getHigh());
		}
		step.setValue(nv.getRange().getStep());
	}

}

/**
 * A ListModel view of this object. Editing is not supported.
 *
 */
class MyListModel extends DefaultListenable<Event>
		implements ListModel<NumberModel> {

	private NumberValueSetModel numvalset;
	/**
	 * A "cache" Map with the actual values from the range. Although
	 * NumberModels, they really should not be edited. This map ensures we
	 * return the SAME NumberModel every time a value from the range is
	 * selected.
	 */
	private final transient Map<BigDecimal, NumberModel> itemsCache = new HashMap<>();

	public MyListModel(NumberValueSetModel valset) {
		this.numvalset = valset;
		valset.addListener((e) -> notifyListeners(e));
	}

	@Override
	public NumberModel get(int n) {
		BigDecimal val = numvalset.getCurrentValue().getRange().get(n);
		if (!itemsCache.containsKey(val))
			itemsCache.put(val, new NumberModel(val, numvalset.log));
		return itemsCache.get(val);
	}

	@Override
	public int getSize() {
		return numvalset.getCurrentValue().getRange().size().intValue();
	}

	@Override
	public void remove(NumberModel value) {
		throw new UnsupportedOperationException();
	}

	@Override
	public void add(NumberModel value) {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean contains(Object o) {
		return o instanceof NumberModel && numvalset.getCurrentValue()
				.getRange().contains(((NumberModel) o).getValue());
	}

}
